#
# AmadeusGPT
# © M.W. Mathis Labs
# https://github.com/AdaptiveMotorControlLab/AmadeusGPT
# Please see AUTHORS for contributors.
#
# Licensed under Apache 2.0

__version__ = "0.1.1rc2"
VERSION = __version__
